﻿Imports System.IO
Imports System.Data.SqlClient
Imports System.Net
Imports System.Text
Imports System.Runtime.InteropServices
Imports Newtonsoft.Json


' enthält alle Daten zu einem einzelnen Record as DataSet
Public Class drAPI

   ' all import management data
   Public Structure importControl

      Dim ErrorMessage As String
      Dim ConnectionString As String

      Dim StartAtRecord As Integer
      Dim MaxRecordsToProcess As Integer
      Dim ProtocolInputRecord As Boolean
      Dim ExtentedProtocol As Boolean

      Dim NbrOfProductsInFile As Integer
      Dim NbrOfProductsProcessed As Integer
      Dim NbrOfProductsError As Integer
      Dim NbrOfProductsWarnings As Integer
      Dim NbrOfProductsInformation As Integer

   End Structure

   ' product information
   Public RecordIdentifier As String
   Private Category As String
   Private ListPriceNetto As Decimal
   Private ListPriceGros As Decimal
   Private TaxPercent As Decimal
   Private CostPrice As Decimal
   Private StockAvailable As Integer
   Private ProductWeight As Integer
   Private Currency As String
   Private PriceType As String
   Private DeliveryTime As String
   Private Rabatt As Decimal
   Private ProductFolder As String
   Public jsonOutput As String

   ' status information
   Public isOK As Boolean
   Public hasWarnings As Boolean
   Public hasInformation As Boolean
   Public warningMessage As String
   Public infoMessage As String
   Public testMode As Boolean = False

   ' save connection received in New()
   Dim conn As SqlConnection

   ' manage input data
   Private products As DataTable = Nothing
   Public nextProductToProcess As Integer
   Public maxProductToProcess As Integer

   ' will be TRUE in new() if all is ok
   Public isInitialized As Boolean = False

   ' error message if isOK = false or function exits with false
   Public errorMessage As String

   ' DreamRobot credentials and properties
   Private loginURL As String = "https://api.dreamrobot.de/rest/v1.0/token.php"
   Private requestURL As String = "https://api.dreamrobot.de/rest/v1.0/stock/"

   ' add userID and PWD for DreamRobot API Access here
   Private userID As String = "XXXXXXXXX"
   Private userPW As String = "XXXXXXXXX"

   Private userAccessToken As String = String.Empty
   Private userAccessTokenExpiry As DateTime = Nothing

   Private stockData As New drStockData
   Private drAPIResult As drReturnData


   ' ****************************************************************************************************
   ' *** New() function to create and initialize class
   ' ****************************************************************************************************

   Public Sub New(ByRef import As importControl)

      Dim sql As String = String.Empty
      Dim dset As New DataSet

      ' indicate "not yet initialized"
      isInitialized = False
      errorMessage = String.Empty


      ' create connection
      ' -------------------------------------------------------------------------------------
      Try
         conn = New SqlConnection(import.ConnectionString)
         conn.Open()
      Catch ex As Exception
         errorMessage = "in drAPI.New (1): " & errorInfo(ex)
         Exit Sub
      End Try


      ' read changed products into dataset
      ' -------------------------------------------------------------------------------------
      Try

         ' ***********************************************************************************************************************************
         ' Das Interface geht von einer Datenbanktabelle mit der folgenden Struktur aus:

         ' CREATE TABLE [dbo].[ksAmazonListing](
         '   [ProductID] [varchar](20) Not NULL,
         '   [ProductIDType] [smallint] Not NULL,
         '   [ProductReference] [varchar](20) Not NULL,
         '   [DistinctiveTitle] [varchar](250) Not NULL,
         '   [ProductAvailability] [smallint] Not NULL,
         '   [PriceAmount] [Decimal](9, 2) Not NULL,
         '   [isUVP] [bit] Not NULL,
         '   [TaxPercent] [Decimal](3, 1) Not NULL,
         '   [StockOnHand] [Int] Not NULL,
         '   [ProductSize] [varchar](50) NULL,
         '   [ProductWeight] [Int] NULL,
         '   [Infos] [varchar](50) NULL,
         '   [Rabatt] [Decimal](6, 2) Not NULL,
         '   [ItemGroup] [varchar](20) Not NULL,
         '   [isAmazon] [bit] Not NULL,
         '   [FirstUpload] [DateTime] Not NULL,
         '   [LastUpdate] [DateTime] Not NULL,
         '   [ListingAction] [smallint] Not NULL,
         '   Constraint [PK_ksAmazonListing] PRIMARY KEY CLUSTERED ([ProductID] Asc)
         '              With (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, IGNORE_DUP_KEY = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) ON [PRIMARY]
         ') ON [PRIMARY]

         ' Gelesen werden alle Zeilen mit ListingAction = 2, diese wurden zuvor individuell markiert als relevant für Update via API
         ' DRUpdater.exe bearbeitet alle Zeilen der Trefferliste, die hier in drAPI.new() gelesen und in einer DataTable gespeichert wird.
         ' ***********************************************************************************************************************************

         sql = "select * from ksAmazonListing where ListingAction = 2"
         Using dapt As New SqlDataAdapter(sql, conn)
            dapt.MissingSchemaAction = MissingSchemaAction.Add
            dapt.Fill(dset, "product")
         End Using
         conn.Close()

         ' store all rows in datatable "products" to process them line by line
         products = dset.Tables("product").Copy()
         import.NbrOfProductsInFile = products.Rows.Count
         nextProductToProcess = 0
         dset.Dispose()

         ' setup counters to process records
         If import.StartAtRecord > 1 Then nextProductToProcess = import.StartAtRecord
         If import.MaxRecordsToProcess > 0 Then
            maxProductToProcess = import.MaxRecordsToProcess
         Else
            maxProductToProcess = import.NbrOfProductsInFile
         End If

      Catch ex As Exception
         errorMessage = "in drAPI.New (2): " & errorInfo(ex)
         Exit Sub
      End Try


      ' initialize dream robot API
      ' -------------------------------------------------------------------------------------
      If Not getAccessToken() Then
         Exit Sub
      End If


      ' return after all is right
      ' -------------------------------------------------------------------------------------
      isInitialized = True

   End Sub


   ' ****************************************************************************************************
   ' *** dream robot connectivity functions
   ' ****************************************************************************************************

   ' receive initial access token
   Public Function getAccessToken() As Boolean

      Dim data As String = "{""grant_type"": ""client_credentials""}"
      Dim postdata As Byte() = Encoding.UTF8.GetBytes(data)

      'Token-Header:  "Authorization: Basic {Benutzer;Doppelpunkt;Passwort -> als Base64}"
      'curl_setopt($ch, CURLOPT_HTTPHEADER, Array("Authorization : Basic ".$encodedAuth));
      'req.Credentials = New NetworkCredential(userID, userPW)

      Try

         Dim req As HttpWebRequest = DirectCast(HttpWebRequest.Create(loginURL), HttpWebRequest)
         req.Method = "POST"
         req.ContentType = "application/json;encoding=utf-8"
         req.ContentLength = postdata.Length
         req.Accept = "application/json"
         req.Timeout = 600000
         req.Headers.Add("Authorization", "Basic " & Convert.ToBase64String(Encoding.UTF8.GetBytes(userID & ":" & userPW), Base64FormattingOptions.None))

         Dim _stream As Stream = req.GetRequestStream()
         _stream.Write(postdata, 0, postdata.Length)
         _stream.Close()

         Dim _response As HttpWebResponse = req.GetResponse()
         Dim _reader As New StreamReader(_response.GetResponseStream)
         data = _reader.ReadToEnd
         _reader.Close()
         _response.Close()

         ' now have the access token plus expire time
         Dim results As drLoginData = JsonConvert.DeserializeObject(Of drLoginData)(data)
         If results.ack = "success" Then
            userAccessToken = results.access_token
            userAccessTokenExpiry = Now.AddSeconds(results.expires_in - 1)
         Else
            drAPIResult = JsonConvert.DeserializeObject(Of drReturnData)(prepareReturnData(data))
            errorMessage = "Fehler in drAPI.getAccessToken(): " & vbCrLf &
                        "http Returncode: " & drAPIResult.code & vbCrLf &
                        "Error: " & drAPIResult._error & vbCrLf &
                        "Description: " & drAPIResult.description
         End If

      Catch ex As webException
         Dim _response As HttpWebResponse = ex.Response
         Dim _reader As New StreamReader(_response.GetResponseStream)
         data = _reader.ReadToEnd
         drAPIResult = JsonConvert.DeserializeObject(Of drReturnData)(prepareReturnData(data))
         errorMessage = "in drAPI.getAccessToken(): " & errorInfo(ex) & vbCrLf &
                        "http Returncode: " & drAPIResult.code & vbCrLf &
                        "Error: " & drAPIResult._error & vbCrLf &
                        "Description: " & drAPIResult.description
         Return False
      End Try

      Return True

   End Function

   ' read the next record and provide column information
   Public Function prepareProduct() As Boolean

      ' *************************************************************************************************************************
      ' Dream Robot requests:
      '                          Methode    Anfrage                 Antwort
      '                                     Feldname Datentyp       Feldname Datentyp

      ' stock/stock_id}/         GET        select   Select_type    stock    Stock_type Einen Lagerartikel abrufen.
      '                          POST       stock    Stock_type     stock_id integer    Einen Lagerartikel überarbeiten.
      '                          DELETE                                                 Einen Lagerartikel löschen.
      ' stock/{stock_id}/cia/    POST                               stock_id integer    Für den Lagerartikel einen PüMa triggern
      ' *************************************************************************************************************************

      isOK = True

      Try
         ' get product zu process
         Dim current As DataRow = getCurrentProduct()
         If IsNothing(current) Then
            Throw New Exception("Kein aktuelles Produkt gefunden")
         End If

         ' and move to next product
         nextProductToProcess += 1

         ' provide product details for easy access
         RecordIdentifier = current("ProductID")
         ListPriceGros = current("PriceAmount").ToString.Replace(".", ",")
         PriceType = IIf(Convert.ToBoolean(current("isUVP")), "UVP", "GLP")
         TaxPercent = current("TaxPercent").ToString.Replace(".", ",")
         StockAvailable = current("StockOnHand")
         ProductWeight = current("ProductWeight")
         Rabatt = current("Rabatt").ToString.Replace(".", ",")
         If ProductWeight > 920 Then
            ProductFolder = "/bisMax"
         ElseIf ProductWeight > 420 Then
            ProductFolder = "/bis920"
         Else
            ProductFolder = "/bis420"
         End If

         ' build object structure as base for json upload
         Dim amazonData As New drAmazonData
         amazonData.sku = current("ProductReference")
         amazonData.quantity = StockAvailable
         amazonData.price = Format(ListPriceGros, "0.00")
         amazonData.prime_template = current("ItemGroup")

         Dim freetextData As New List(Of drFreetext)
         freetextData.Add(New drFreetext With {.id = 1, .value = Format(Rabatt, "0.00")})
         freetextData.Add(New drFreetext With {.id = 2, .value = ProductWeight.ToString})
         freetextData.Add(New drFreetext With {.id = 3, .value = PriceType})

         Dim uplData As New drItemData
         uplData.id = "SKU_" & drData.RecordIdentifier
         uplData.folder = ProductFolder
         uplData.price = Format(ListPriceGros, "0.00")
         uplData.tax = Format(TaxPercent, "0.0")
         uplData.quantity = StockAvailable
         uplData.weight = ProductWeight

         uplData.freetext = freetextData
         uplData.amazon = amazonData

         stockData.stock = uplData

         jsonOutput = JsonConvert.SerializeObject(stockData)

      Catch ex As Exception
         errorMessage = "in drAPI.prepareProduct: " & errorInfo(ex)
         isOK = False
         Return False
      End Try

      Return True

   End Function

   ' update product using DR API
   Public Function processProduct() As Boolean

      ' if token is about to expire, request a new one
      If DateDiff(DateInterval.Second, Now(), userAccessTokenExpiry) < 0 Then
         getAccessToken()
      End If


      ' jsonOutput has been prepared in prepareProduct()
      Dim stockURL As String = requestURL & "SKU_" & drData.RecordIdentifier & "/"
      Dim postdata As Byte() = Encoding.UTF8.GetBytes(jsonOutput)
      Dim data As String

      isOK = True
      Dim req As HttpWebRequest = DirectCast(HttpWebRequest.Create(stockURL), HttpWebRequest)


      ' *****************************************************************
      ' no updating of real products if in test mode
      If testMode Then Return True
      ' *****************************************************************

      Try

         req.Method = "POST"
         req.ContentType = "application/json;encoding=utf-8"
         req.ContentLength = postdata.Length
         req.Accept = "application/json"
         req.Timeout = 600000
         req.Headers.Add("Authorization: Bearer " & userAccessToken)

         ' send stock data
         Dim _stream As Stream = req.GetRequestStream()
         _stream.Write(postdata, 0, postdata.Length)
         _stream.Close()

         ' retrieve response
         Dim _response As HttpWebResponse = req.GetResponse()
         Dim _reader As New StreamReader(_response.GetResponseStream)
         data = _reader.ReadToEnd
         _reader.Close()
         _response.Close()

         ' check result
         drAPIResult = JsonConvert.DeserializeObject(Of drReturnData)(prepareReturnData(data))
         If Not drAPIResult.ack = "success" Then
            ' error updating item
            errorMessage = "Fehler bei Aktualisierung Produkt in DreamRobot : " & vbCrLf &
                           "Acknowledge: " & drAPIResult.ack & vbCrLf &
                           "http Returncode: " & drAPIResult.code & vbCrLf &
                           "Error: " & drAPIResult._error & vbCrLf &
                           "Description: " & drAPIResult.description
            isOK = False
         End If

      Catch ex As WebException
         Dim _response As HttpWebResponse = ex.Response
         Dim _reader As New StreamReader(_response.GetResponseStream)
         data = _reader.ReadToEnd
         drAPIResult = JsonConvert.DeserializeObject(Of drReturnData)(prepareReturnData(data))
         errorMessage = "in drAPI.processProduct(): " & errorInfo(ex) & vbCrLf &
                        "Acknowledge: " & drAPIResult.ack & vbCrLf &
                        "http Returncode: " & drAPIResult.code & vbCrLf &
                        "Error: " & drAPIResult._error & vbCrLf &
                        "Description: " & drAPIResult.description
         isOK = False
         Return False
      End Try

      Return True

   End Function

   ' trigger püma for a processed product
   Public Function triggerCIA() As Boolean

      ' if token is about to expire, request a new one
      If DateDiff(DateInterval.Second, Now(), userAccessTokenExpiry) < 0 Then
         getAccessToken()
      End If

      Dim ciaURL As String = requestURL & "SKU_" & drData.RecordIdentifier & "/cia/"
      Dim data As String = "{}"
      Dim postdata As Byte() = Encoding.UTF8.GetBytes(data)

      isOK = True
      Dim req As HttpWebRequest = DirectCast(HttpWebRequest.Create(ciaURL), HttpWebRequest)


      ' *****************************************************************
      ' no updating of real products if in test mode
      If testMode Then Return True
      ' *****************************************************************

      Try

         req.Method = "POST"
         req.ContentType = "application/json;encoding=utf-8"
         req.ContentLength = postdata.Length
         req.Accept = "application/json"
         req.Timeout = 600000
         req.Headers.Add("Authorization: Bearer " & userAccessToken)

         ' send empty data
         Dim _stream As Stream = req.GetRequestStream()
         _stream.Write(postdata, 0, postdata.Length)
         _stream.Close()

         ' retrieve response
         Dim _response As HttpWebResponse = req.GetResponse()
         Dim _reader As New StreamReader(_response.GetResponseStream)
         data = _reader.ReadToEnd
         _reader.Close()
         _response.Close()

         ' check result
         drAPIResult = JsonConvert.DeserializeObject(Of drReturnData)(prepareReturnData(data))
         If Not drAPIResult.ack = "success" Then
            ' error updating cia
            errorMessage = "Fehler beim triggern Bestandsabgleich in DreamRobot : " & vbCrLf &
                           "Acknowledge: " & drAPIResult.ack & vbCrLf &
                           "http Returncode: " & drAPIResult.code & vbCrLf &
                           "Error: " & drAPIResult._error & vbCrLf &
                           "Description: " & drAPIResult.description
            isOK = False
         End If

      Catch ex As WebException
         Dim _response As HttpWebResponse = ex.Response
         Dim _reader As New StreamReader(_response.GetResponseStream)
         data = _reader.ReadToEnd
         drAPIResult = JsonConvert.DeserializeObject(Of drReturnData)(prepareReturnData(data))
         errorMessage = "in drAPI.triggerCIA(): " & errorInfo(ex) & vbCrLf &
                        "Acknowledge: " & drAPIResult.ack & vbCrLf &
                        "http Returncode: " & drAPIResult.code & vbCrLf &
                        "Error: " & drAPIResult._error & vbCrLf &
                        "Description: " & drAPIResult.description
         isOK = False
         Return False
      End Try

      Return True

   End Function


   ' ****************************************************************************************************
   ' *** help functions
   ' ****************************************************************************************************

   ' prepare a response in case of error for readability
   Private Function prepareReturnData(ByRef data As String) As String

      If String.IsNullOrEmpty(data) Then Return String.Empty

      Dim retdata As String = data
      If retdata.StartsWith("{""error"":[{") Then retdata = Right(retdata, retdata.Length - 10).Replace("}]", "")
      retdata = retdata.Replace("""error"":", """_error"":")

      Return retdata

   End Function

   ' replacement value for database columns of string types
   Public Shared Function stringDBNull(ByRef value As Object, ByRef newval As String) As String

      If IsDBNull(value) OrElse IsNothing(value) Then
         Return newval
      Else
         Return IIf(value.ToString = "", newval, value.ToString)
      End If

   End Function


   ' ****************************************************************************************************
   ' *** properties
   ' ****************************************************************************************************

   ' access to current datarow
   Public ReadOnly Property getCurrentProduct() As DataRow

      Get
         If Not isInitialized Then Return Nothing
         If IsNothing(products) Then Return Nothing
         If nextProductToProcess >= products.Rows.Count Then Return Nothing
         Return products(nextProductToProcess)
      End Get

   End Property


   '****************************************************************************************************
   ' error management
   '****************************************************************************************************

   Public Shared Function errorInfo(ByRef ex As Exception) As String

      Dim errText As String = String.Empty

      If IsNothing(ex) Then Return String.Empty

      errText &= ex.Message
      ' change to ex.HResult.ToString from .NET 4.5
      errText &= " [" & Marshal.GetHRForException(ex).ToString & "]"
      If Not IsNothing(ex.InnerException) Then errText &= " / " & ex.InnerException.Message

      Return errText

   End Function

End Class


' ****************************************************************************************************
' *** Dreamrobot communication structures
' ****************************************************************************************************

' Dream Robot Authentification structure
Public Class drLoginData

   ' this is the return data from the token request to DreamRobot
   '{
   ' "access_token":"0891b96063d76509fbe48c6a683a9f39e948844b4e4193b0800403508362a43b3039bcc4405a1858a4edddad3f30c6b2860f9a75668bd23e4a37d51c61110d5ac6cc186d44e1da770fda2937f21f0ecb",
   ' "expires_in":3600,
   ' "token_type":"Bearer",
   ' "scope":"*",
   ' "ack":"success"
   '}

   Public Property access_token As String
   Public Property expires_in As Integer
   Public Property token_type As String
   Public Property scope As String
   Public Property ack As String

End Class

' Dream Robot return structure for errors and success
Public Class drReturnData

   ' this is the response data by DreamRobot after an error
   '{
   ' "error":[
   '          {
   '           "error":"Bad Request",
   '           "description":"Unknown input-field \"POST https:\/\/api.dreamrobot.de\/rest\/v1.0\/stock\/SKU_9780857056429\/ stock[0]\" with value \"Array\".",
   '           "code":400
   '          }
   '         ],
   ' "ack":"error"
   '}

   ' '{"error":[{' at the beginning and the '}]' before "ack" are removed to create a one level return structure deserialized herein

   ' this is the response after successful item update:
   ' {"stock_id":700436,"ack":"success"}

   Public Property stock_id As Long
   Public Property code As String
   Public Property _error As String
   Public Property description As String
   Public Property ack As String

End Class

' Dream Robot top level structure
Public Class drStockData
   Public Property stock As drItemData
End Class

' Dream Robot item structure
Public Class drItemData
   Public Property id As String
   Public Property price As Decimal
   Public Property tax As Decimal
   Public Property weight As Integer
   Public Property folder As String
   Public Property quantity As Integer
   Public Property freetext() As List(Of drFreetext)
   Public Property amazon As drAmazonData
End Class

' Dream Robot Amazon item
Public Class drAmazonData
   Public Property sku As String
   Public Property quantity As Integer
   Public Property price As Decimal
   Public Property prime_template As String
End Class

' Dream Robot freetext structure
Public Class drFreetext
   Public Property id As Integer
   Public Property value As String
End Class
